/*-----------------------------------------------------------------------

File  : semantic.h

Author: Stephan Schulz (schulz@eprover.org)

Contents
 
  Semantic checks and annotations for nanoLang.

  Copyright 2015 by the author.
  This code is released under the GNU General Public Licence.

Changes

<1> Mon Apr 20 01:12:05 CEST 2015
    New

-----------------------------------------------------------------------*/

#include "semantic.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: get_type_ast_type()
//
//   Given an AST node corresponding to a basic type, return the
//   corresponding type. Otherwise, exit with error.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static TypeIndex get_type_ast_type(AST_p ast)
{
   TypeIndex res = 0;

   switch(ast->type)
   {
   case t_STRING:
         res = T_String;
         break;
   case t_INTEGER:
         res = T_Integer;
         break;
   default:
         assert(false && "Unexpected AST type in get_type_ast_type()");
         break;
   }
   return res;         
}

/*-----------------------------------------------------------------------
//
// Function: get_param_types()
//
//  Add the types of the parameters represented by ast to ntype.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void get_param_types(NanoType_p ntype, AST_p ast)
{
   if(ast)
   {
      switch(ast->type)
      {
      case params:
            get_param_types(ntype, ast->child1);
            break;
      case paramlist:
            get_param_types(ntype, ast->child1);
            get_param_types(ntype, ast->child2);
            break;
      case param:
            ntype->typeargs[ntype->typeargno++] =
               get_type_ast_type(ast->child1);
            break;
      default:
            assert(false && "Unexpected AST type in get_param_types()");
            break;
      }
   }
   
}


/*-----------------------------------------------------------------------
//
// Function: st_insert_params()
//
//   Insert the formal parameter list pointed to by ast into the
//   symbol table. Return success (true) or failure.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

static bool st_insert_params(SymbolTable_p st, TypeTable_p tt, AST_p ast)
{
   bool res = true;

   if(ast)
   {
      switch(ast->type)
      {
      case params:
            res = st_insert_params(st, tt, ast->child1) && res;
            break;
      case paramlist:
            res = st_insert_params(st, tt, ast->child1) && res;
            res = st_insert_params(st, tt, ast->child2) && res;
            break;
      case param:
            res = STInsertVarDef(st, tt, ast) && res;
            break;
      default:
            assert(false && "Unexpected AST type in get_param_types()");
            break;
      }
   }
   return res;
}



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: STInsertVarDef()
//
//   Insert a variable definition into the given symbol table.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/
bool STInsertVarDef(SymbolTable_p st, TypeTable_p tt, AST_p def)
{
   TypeIndex type;
   assert((def->type == vardef) || (def->type == param));
   
   type = get_type_ast_type(def->child1);
   return STInsertSymbol(st, def->child2->litval, type, 
                         def->child1->line, def->child1->col);
}

/*-----------------------------------------------------------------------
//
// Function: STInsertFunDef()
//
//   Insert a function definition into the given symbol table (and
//   ensure that the type is in the type table).
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

bool STInsertFunDef(SymbolTable_p st, TypeTable_p tt, AST_p def)
{
   NanoTypeCell ntype;
   TypeIndex    type;
   
   assert(def->type == fundef);

   ntype.constructor = tc_function;
   ntype.typeargno   = 1;
   ntype.typeargs[0] = get_type_ast_type(def->child1);
   
   get_param_types(&ntype, def->child3);
   type = TypeTableGetTypeIndex(tt, &ntype);
   return STInsertSymbol(st, def->child2->litval, type, 
                         def->child1->line, def->child1->col);   
}


/*-----------------------------------------------------------------------
//
// Function: STBuildAllTables()
//
//   Traverse the AST, build all symbol tables and annotate all AST
//   relevant AST nodes with the symbol table valid at that
//   node. Return true on success, false otherwise.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool STBuildAllTables(SymbolTable_p st, TypeTable_p tt, AST_p ast)
{
   bool res = true;

   assert(st);

   if(ast)
   {
      /* printf("\nSymboltable at node: ");
      ASTCellPrint(ast);
      SymbolTablePrintGlobal(stdout, st, tt); */

      ast->context = st;
      switch(ast->type)
      {
      case vardef:
            res = STBuildAllTables(st, tt, ast->child1) && res;
            res = STBuildAllTables(st, tt, ast->child2) && res;
            res = STInsertVarDef(st, tt, ast) && res;
            break;
      case fundef:
            res = STInsertFunDef(st, tt, ast) && res; 
            res = STBuildAllTables(st, tt, ast->child1) && res;            
            res = STBuildAllTables(st, tt, ast->child2) && res;            
            st = STEnterContext(st);
            res = st_insert_params(st, tt, ast->child3) && res;
            res = STBuildAllTables(st, tt, ast->child3) && res;
            res = STBuildAllTables(st, tt, ast->child4) && res;
            st = STLeaveContext(st);
            break;
      case body:
            st = STEnterContext(st);
            res = STBuildAllTables(st, tt, ast->child1) && res;
            res = STBuildAllTables(st, tt, ast->child2) && res;
            st = STLeaveContext(st);            
            break;
      default:
            res = STBuildAllTables(st, tt, ast->child1) && res;
            res = STBuildAllTables(st, tt, ast->child2) && res;
            res = STBuildAllTables(st, tt, ast->child3) && res;
            res = STBuildAllTables(st, tt, ast->child4) && res;
            break;
      }
   }
   return res;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


